//
// Created by welsch on 25.06.18.
//

#include "TemplateFitter.h"

#include "RooArgSet.h"
#include "RooMCStudy.h"


#include "TF1.h"
#include "TCanvas.h"


TemplateFitter::TemplateFitter(std::shared_ptr<RooRealVar> observable,
                               const std::string &sig,
                               const std::vector<std::string> &vbkg,
                               const std::map<std::string,
                                       std::shared_ptr<RooDataSet>> &mds) :
                                m_observable{observable},m_sig{sig},
                                m_vbkg{vbkg}, m_mds{mds}  {
    create_templates();
    create_model();
}


void TemplateFitter::create_templates() {
    std::cout << "Create templates:" << std::endl;

    for(const auto &tmds : m_mds) {
        auto temp= std::make_shared<RooHistPdf>(tmds.first.c_str(), tmds.first.c_str(),
                *m_observable, *(tmds.second->binnedClone()));
        m_mtemp.emplace(tmds.first, temp);
        m_mexp_yields.emplace(tmds.first, tmds.second->sumEntries());
        std::cout << "\t" << tmds.first << " template created." << std::endl;
    }
}

void TemplateFitter::create_model() {
    std::cout << "Create fit model." << std::endl;


    auto nsig = std::make_shared<RooRealVar>(("N"+m_sig).c_str(),
            (m_sig + " Yield").c_str(), 0, 10000);
    m_myields.emplace(m_sig, nsig);
    RooArgSet syields(*nsig);
    RooArgSet stemps(*m_mtemp[m_sig]);
    for(const auto & bkg : m_vbkg) {
        auto nbkg = std::make_shared<RooRealVar>(("N"+bkg).c_str(),
                                                 (bkg + " Yield").c_str(), 0, 10000);
        m_myields.emplace(bkg, nbkg);
        syields.add(*nbkg);
    }

    for(auto temp : m_mtemp) stemps.add(*temp.second);
    auto model = std::make_shared<RooAddPdf>("template model", "Template model",
            stemps, syields);
    m_model = model;

}

void TemplateFitter::reset_yields() {
    for(auto pair : m_myields){
        pair.second->setVal(m_mexp_yields[pair.first]);
    }
}

void TemplateFitter::set_yield(std::string sample, double yield) {
    m_myields[sample]->setVal(yield);
}

void TemplateFitter::do_linearity_check(double from, double to, int nsteps) {
    std::cout << "Perform a linearity check for the signal yield from "
                 << from << " to " << to << std::endl;

    double step_size = (to - from)/nsteps;
    std::vector<double> vfitted_yield;
    std::vector<double> vfitted_yield_error;
    std::vector<double> vnominal_yield;
    std::vector<std::pair<double, double>> vmean_error;

    for(double yield = from; yield <= to; yield += step_size) {
        std::cout << "\tTesting yield: " << yield << std::endl;

        auto mean_error = do_toys(yield, 100);
        vfitted_yield.push_back(mean_error.first);
        vfitted_yield_error.push_back(mean_error.second);
        vnominal_yield.push_back(yield);
        vmean_error.push_back(mean_error);

    }

    for(const auto &mean_error : vmean_error) {
        std::cout << "\tMean br: " << mean_error.first <<
                  " +- " << mean_error.second << std::endl;
    }

    std::vector<double> zeros(vnominal_yield.size(), 0.0);

    TGraphErrors linear(vfitted_yield.size(), vfitted_yield.data(),
                        vnominal_yield.data(), zeros.data(),
                        vfitted_yield_error.data());
    fit_linearity_check(linear);
}

void TemplateFitter::fit_linearity_check(TGraphErrors linear){
    TF1 linearfunction("linear function", "[0] + [1]*x");
    linearfunction.SetParNames("offset", "slope");
    linearfunction.SetParameters(0,1);
    linearfunction.SetLineColor(kRed);

    std::cout << "Fit linear function to linearity check:" << std::endl;
    linear.Fit(&linearfunction);
    linear.SetMarkerStyle(20);
    TCanvas canv("canv_linearity", "canv_linearity", 800, 600);

    linear.Draw("AP");
    linearfunction.Draw("SAME");
    linear.SetTitle(" ;Input Signal Yield;Fitted Signal Yield");

    canv.SaveAs("linearity_check.pdf", "pdf");
}

std::pair<double, double> TemplateFitter::do_toys(double signal_yield,
                                                  int ntoys) {
    reset_yields();
    set_yield(m_sig, signal_yield);

    RooMCStudy mcStudy(
            *m_model,
            *m_observable,
            RooFit::Binned(false),
            RooFit::Extended(true),
            RooFit::FitOptions(RooFit::Save(true)),
            RooFit::NumCPU(4),
            RooFit::Silence(kTRUE));
    mcStudy.generateAndFit(ntoys);

    auto fit_param_set = mcStudy.fitParDataSet();
    auto  meanVar = fit_param_set.meanVar(*m_myields[m_sig]);
    return std::make_pair<double, double>(meanVar->getVal(), meanVar->getError());
};